/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2007, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

 ////////                 
   //    //    ////////    This file is part of rqal, a.k.a. _Richard's Quality Library_ .
   //    //   //    //     Visit http://www.ranton.org for more information.
   //////     //    //     Email ranton@ranton.org for support or licensing.  
   //   //     //// //     
   //    //          //    Copyright (c) 2007 Richard Neal Anton
   //    //          //    All Rights Reserved.
  ///    //           //

#ifndef __RQ__
#include "rq.h"
#endif

// NOTE: reimpliment this using fstream instead of antiquated c stdlib FILE * calls

#include <errno.h>
#include <cstring>

using namespace std;

RqFile::RqFile(void)
{
	mDesc = 0;
}

RqFile::~RqFile()
{
	mDesc = 0;
}

// Open a file for reading or writing, we don't support both at once.
void RqFile::OpenForRead(const char *filename)
{
	RqAssert(filename);
	mDesc = fopen(filename,"rb");
	if(!mDesc)
	{
		int err = errno;
		RqLog(RqLog_ERROR,"Error # %d opening %s for reading",err,filename);
		RqThrow(RqError::OPEN_ERR,"Unable to open file");
	}
}

// This will truncate the file if true, if false will not
// truncate, but position in file is undefined so call Seek
void RqFile::OpenForWriting(const char *filename, bool appendFlag)
{
	RqAssert(filename);
	
	if( !appendFlag )
		mDesc = fopen(filename,"wb");
	else
		mDesc = fopen(filename,"rb+");
	if( !mDesc )
	{
		int err = ferror(mDesc);
		RqLog(RqLog_ERROR,"Error %d opening %s for writing",err,filename);
		if( err == EACCES ) // this is what we get for no write permission right?
			RqThrow(RqError::PERM_ERR,"Error writing file: permission denied");
		else if( err == EINVAL || err == EISDIR || err == ENOTDIR ) // NOTE: what do we get for pathname problems ??
			RqThrow(RqError::PATH_ERR,"Error writing file: Invalid file path");
		else
			RqThrow(RqError::UNKNOWN,"Error opening file");
	}
}

// Close the current file
void RqFile::Close(void)
{
	if( fclose(mDesc) )
	{
		int err = ferror(mDesc);
		RqLog(RqLog_ERROR,"Error %d closing file",err);
		RqThrow(RqError::CLOSE_ERR,"Error closing file");
	}
	mDesc = 0;
}

// Tell current position in file
RqFilePos RqFile::Tell(void)
{
	long ret = ftell( mDesc );
	if( ret < 0 )
	{
		int err = ferror(mDesc);
		RqLog(RqLog_ERROR,"Error %d from file tell",err);
		RqThrow(RqError::TELL_ERR,"Error getting file position");
	}
	return (RqFilePos)ret;
}

// Seek to a new position in the file
void RqFile::Seek(RqSeekFrom from, RqFilePos pos)
{
	int whence;
	if(from == RqSeekPos )
		whence = SEEK_CUR;
	else if(from == RqSeekStart )
		whence = SEEK_SET;
	else // pos == RqSeekEnd
		whence = SEEK_END;

	int ret = fseek(mDesc,pos,whence);
	if( ret < 0 )
	{
		int err = ferror(mDesc);
		RqLog(RqLog_ERROR,"Error %d from file seek",err);
		RqThrow(RqError::SEEK_ERR,"Error setting file position");
	}
}

// Read from the file, if EOF may read fewer bytes than requested or none,
// Return's number of bytes read.
RqFilePos RqFile::Read(void *buf, RqFilePos len)
{
	RqAssert(buf && len);
	size_t ret;
	
	ret = fread( buf, 1, len, mDesc );
	if( ret < 0 )
	{
		int err = ferror(mDesc);
		RqLog(RqLog_ERROR,"Error %d reading from file",err);
		RqThrow(RqError::READ_ERR,"Error reading from file");
	}

	return (RqFilePos)ret;
}

// Write to file
void RqFile::Write(const void *buf, RqFilePos len)
{
	RqAssert(buf && len);
	size_t ret = fwrite(buf,1,len,mDesc);
	if( ret != (size_t)len )
	{
		int err = ferror(mDesc);
		RqLog(RqLog_ERROR,"Error %d writing to file",err);
		if( err == ENOSPC || !err )
			RqThrow(RqError::DISK_FULL,"Not enough disk space");
		RqThrow(RqError::WRITE_ERR,"Error writing to file");
	}
}






// This reads a byte from the file.
RqUns8 RqFile::ReadByte(void)
{
	RqUns8 theData[1];
	Read(theData, 1);
	return theData[0];
}

// These read a word from the file.
// Big-endian
RqUns16 RqFile::ReadWordBE(void)
{
	RqUns8 theData[2];
	Read((void *)theData, 2);
	RqUns16 retWord = theData[0];
	retWord <<= 8;
	retWord |= theData[1];
	return retWord;
}
// Little-endian
RqUns16 RqFile::ReadWordLE(void)
{
	RqUns8 theData[2];
	Read((void *)theData, 2);
	RqUns16 retWord = theData[1];
	retWord <<= 8;
	retWord |= theData[0];
	return retWord;
}

// These read a long from the file.
// Big-endian
RqUns32 RqFile::ReadLongBE(void)
{
	RqUns8 theData[4];
	Read((void *)theData,4);
	RqUns32 retLong = theData[0];
	retLong <<= 8;
	retLong |= theData[1];
	retLong <<= 8;
	retLong |= theData[2];
	retLong <<= 8;
	retLong |= theData[3];
	return retLong;
}
// Little-endian
RqUns32 RqFile::ReadLongLE(void)
{
	RqUns8 theData[4];
	Read((void *)theData,4);
	RqUns32 retLong = theData[3];
	retLong <<= 8;
	retLong |= theData[2];
	retLong <<= 8;
	retLong |= theData[1];
	retLong <<= 8;
	retLong |= theData[0];
	return retLong;
}

// This writes a byte to the file.
void RqFile::WriteByte(RqUns8 inByte)
{
	Write(&inByte,1);
}

// These write a word to the file.
// Big-endian
void RqFile::WriteWordBE(RqUns16 inWord)
{
	// The write buffer.
	RqUns8 theData[2];
	// Put the bytes in the necessary order.
	theData[0] = inWord >> 8;
	theData[1] = inWord & 0xFF;
	// Write the data.
	Write((void *)theData,2);
}
// Little-endian
void RqFile::WriteWordLE(RqUns16 inWord)
{
	// The write buffer.
	RqUns8 theData[2];
	// Put the bytes in the necessary order.
	theData[0] = inWord & 0xFF;
	theData[1] = inWord >> 8;
	// Write the data.
	Write((void *)theData,2);
}

// These write a long to the file.
// Big-endian
void RqFile::WriteLongBE(RqUns32 inLong)
{
	// The write buffer.
	RqUns8 theData[4];
	// Put the bytes in the necessary order.
	theData[0] = (RqUns8)(inLong >> 24);
	theData[1] = (RqUns8)((inLong >> 16) & 0xFF);
	theData[2] = (RqUns8)((inLong >> 8) & 0xFF);
	theData[3] = (RqUns8)(inLong & 0xFF); 
	// Write the data.
	Write((void *)theData,4);
}
// Little-endian.
void RqFile::WriteLongLE(RqUns32 inLong)
{
	// The write buffer.
	RqUns8 theData[4];
	// Put the bytes in the necessary order.
	theData[0] = (RqUns8)(inLong & 0xFF);
	theData[1] = (RqUns8)((inLong >> 8) & 0xFF);
	theData[2] = (RqUns8)((inLong >> 16) & 0xFF);
	theData[3] = (RqUns8)(inLong >> 24);
	// Write the data.
	Write((void *)theData,4);
}

const char *RqFile::GetFileNameFromFullPath(const char *inpath)
{
	const char *ret = NULL;

	while(*inpath)
	{
		if(*inpath == ':')
			ret = NULL;
		else if(*inpath == '\\')
			ret = NULL;
		else if(!ret)
			ret = inpath;

		inpath++;
	}
	if(!ret)
		ret = inpath + RqStrLen(inpath);
	
	return ret;
}
