/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2007, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

 ////////                 
   //    //    ////////    This file is part of rqal, a.k.a. _Richard's Quality Library_ .
   //    //   //    //     Visit http://www.ranton.org for more information.
   //////     //    //     Email ranton@ranton.org for support or licensing.  
   //   //     //// //     
   //    //          //    Copyright (c) 2007 Richard Neal Anton
   //    //          //    All Rights Reserved.
  ///    //           //

#ifndef __RQ__
#include "rq.h"
#endif

#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstring>
#include <ctime>

using namespace std;

// -- Default logging settings.
const RqLogType RqLog_DEFAULTMASK = 
#if defined(RQ_DEBUG)
	RqLog_ALL;
#else
	RqLog_WARNING | RqLog_ERROR;
#endif

static RqLogType sLogMask = RqLog_DEFAULTMASK;
static bool sLogTimeFlag = false;
static RqTimeCallback sTimeCallback = RqDefaultTimeCallback;
static RqLogCallback sCallback = RqDefaultLogCallback;
static void* sCallbackData = NULL;



// Set log mask which determines which kinds of messages are actually output.
// returns prior value of mask.
RqLogType RqSetLogMask(RqLogType inMask)
{
	RqLogType old = sLogMask;
	sLogMask = inMask;
	return old;
}
#ifdef __cplusplus
bool RqSetLogTime(bool inLogTimeFlag)
{
	bool previous = sLogTimeFlag;
	sLogTimeFlag = inLogTimeFlag;
	return previous;
}
// return value of log time flag.
bool RqGetLogTime(void)
{
	return sLogTimeFlag;
}
#endif

// set callback for getting time value for log entries used if RqGetLogTime() == true
// default returns time(NULL)
RqTimeCallback RqSetTimeCallback( RqTimeCallback inTimeCallback )
{
	RqTimeCallback previous = sTimeCallback;
	sTimeCallback = inTimeCallback;
	return previous;
}

void *RqLogCallbackData(void)
{
	return sCallbackData;
}

// by default assertions, traces, debug msgs, log msgs, warnings, and error messages
// go to stderr, you can override this by setting your output function using this function.

long RqDefaultTimeCallback(void)
{
	return time(NULL);
}

// default message proc that prints to stderr
void RqDefaultLogCallback(RqLogType kind, void *cbData, const char *fmt, va_list ap)
{
	// just pass cout to the stream log callback.
	RqStreamLogCallback(kind,(void *)&std::cout,fmt,ap);
}

static const char *_RqLogPre( RqLogType kind )
{
	const char *pre = "";

	switch(kind)
	{
	case RqLog_TRACE:
		pre = "trace: ";
		break;
	case RqLog_DEBUG:
		pre = "dbg: ";
		break;
	case RqLog_INFO:
		pre = "info: ";
		break;
	case RqLog_MEMORY:
		pre = "mem: ";
		break;
	case RqLog_WARNING:
		pre = "warning: ";
		break;
	case RqLog_ERROR:
		pre = "error: ";
		break;
	case RqLog_ASSERT:
		pre = "assert: ";
		break;
	}
	return pre;
}

// -- Logging output callback for a stream
void RqStreamLogCallback(RqLogType kind, void *cbData, const char *fmt, va_list ap)
{
	char msg[512];
	std::ostream* os = (std::ostream*)cbData;
	
	const char *pre = _RqLogPre(kind);
	
#if defined(RQ_WINDOWS)
	_vsnprintf(msg,511,fmt,ap);
#else
	// ATTN: If your platform doesn't have vsnprintf, it probably has vsprintf but it isn't as safe.
	vsnprintf(msg,511,fmt,ap);
#endif

	if( RqGetLogTime() )
		(*os) << sTimeCallback() << ": ";
	(*os) << pre;
	(*os) << msg;
	(*os) << "\n";
}

// -- Logging output for an fstream that flushes after each message
void RqFileLogCallback(RqLogType kind, void *cbData, const char *fmt, va_list ap)
{
	std::ofstream* file = (std::ofstream*)cbData;
	// call the normal string log handler.
	RqStreamLogCallback(kind,cbData,fmt,ap);
	// flush the file.
	file->flush();
}

static const int kLogCaptureMax = 8096; // 1024;
static int sLogCapturePos = 0;
static char sLogCapture[kLogCaptureMax+1] = { 0 };

static void _RqAddToLogCapture(const char* msg)
{
	int len = RqStrLen(msg); // 1 for newline.
	int remaining = kLogCaptureMax - sLogCapturePos;
	if( len > remaining-1 )
	{
		len = 0;
		if( remaining > 0 )
			len = remaining-1;
	}

	// RqMemCopy can call us, so we must not call it -RNA
	// RqMemCopy(sLogCapture+sLogCapturePos,msg,len);
	{
		char *dst = sLogCapture+sLogCapturePos;
		const char *src = msg;
		int cc = len;
		while(cc--)
			*dst++ = *src++;
	}
	sLogCapturePos += len;
	sLogCapture[sLogCapturePos] = 0;

//	sLogCapture[sLogCapturePos] = '\n';
//	sLogCapture[++sLogCapturePos] = 0;
}

void RqLogCaptureCallback(RqLogType kind, void *cbData, const char *fmt, va_list ap)
{
	char msg[512];
#if defined(RQ_WINDOWS)
	_vsnprintf(msg,511,fmt,ap);
#else
	// ATTN: If your platform doesn't have vsnprintf, it probably has vsprintf but it isn't as safe.
	vsnprintf(msg,511,fmt,ap);
#endif
	
	if( RqGetLogTime() )
	{
		char tbuf[64]; // must hold int as string
		sprintf(tbuf,"%ld: ",sTimeCallback());
		_RqAddToLogCapture(tbuf);
	}
	_RqAddToLogCapture(_RqLogPre(kind));
	_RqAddToLogCapture(msg);
	_RqAddToLogCapture("\n");
}

const char *RqLogCapture(void)
{
	return sLogCapture;
}

void RqLogCaptureReset(void)
{
	sLogCapture[0] = 0;
	sLogCapturePos = 0;
}

// Set output callback for log messages
RqLogCallback RqSetLogCallback( RqLogCallback inCallback, void *cbData )
{
	RqLogCallback oldCallback = sCallback;
	sCallback = inCallback;
	sCallbackData = cbData;
	return oldCallback;
}

// output a message of a specified type
void RqLog(RqLogType kind, const char *fmt, ...)
{
	_RqDoNotify(kind,fmt);
}

// this is just a wrapper to call the proc
void RqLogV(RqLogType kind, const char *fmt, va_list ap)
{
	// if it is in the mask and we have a callback then log the message.
	if( (sLogMask & kind) != 0 && sCallback )
		sCallback(kind,sCallbackData,fmt,ap);
}

void RqDebug(const char *fmt,...) 
{ _RqDoNotify(RqLog_DEBUG, fmt);}

void RqInfo(const char *fmt, ...) 
{ _RqDoNotify(RqLog_INFO, fmt);}

void RqWarning(const char *fmt, ...) 
{ _RqDoNotify(RqLog_WARNING, fmt);}

void RqLogError(const char *fmt, ...) 
{ _RqDoNotify(RqLog_ERROR, fmt);}




RqLogStream::RqLogStream( RqLogType inKind )
: mKind(inKind), mLine(NULL),mPos(0),  mSize(0)
{
}


RqLogStream::~RqLogStream()
{
	// blast the output
	if( mPos > 0 )
	{
		mLine[mPos] = 0;
		RqLog(mKind,"%s",mLine);
	}
	if( mLine )
		delete[] mLine;
}


RqLogStream& RqLogStream::operator<< ( const char *str )
{
	Append(str);
	return *this;
}
	
RqLogStream& RqLogStream::operator<< (const std::string& sobj)
{
	Append(sobj.c_str());
	return *this;
}

RqLogStream& RqLogStream::operator<< (bool val)
{
	Append( val ? "true" : "false" );
	return *this;
}

RqLogStream& RqLogStream::operator<< (char val)
{
	char str[2];
	str[0] = val;
	str[1] = 0;
	Append(str);
	return *this;
}

RqLogStream& RqLogStream::operator<< (int val)
{
	char buf[ kConvertBufSize + 1];
	sprintf(buf, "%d", val);
	Append(buf);
	return *this;
}

RqLogStream& RqLogStream::operator<< (unsigned val)
{
	char buf[ kConvertBufSize + 1];
	sprintf(buf, "%u", val);
	Append(buf);
	return *this;
}

RqLogStream& RqLogStream::operator<< (long val)
{
	char buf[ kConvertBufSize + 1];
	sprintf(buf, "%ld", val);
	Append(buf);
	return *this;
}

RqLogStream& RqLogStream::operator<< (unsigned long val)
{
	char buf[ kConvertBufSize + 1];
	sprintf(buf, "%lu", val);
	Append(buf);
	return *this;
}

RqLogStream& RqLogStream::operator<< (double val)
{
	char buf[ kConvertBufSize + 1];
	sprintf(buf, "%f", val);
	Append(buf);
	return *this;
}

// this does the work.
void RqLogStream::Append(const char *str)
{
	long avail = mSize - mPos;
	long slen = RqStrLen(str);
	if( slen > avail && !Grow( slen - avail ) )
		return; // can't get enough memory, fuck it.

	RqStrCopy(mLine + mPos, str);
	mPos += slen;
}


bool RqLogStream::Grow(long minToGrow)
{
	int minSize = mSize + minToGrow;
	int newSize = mSize * 2;
	if( newSize < minSize )
		newSize = minSize;
	
	char *buf = new char[ newSize+1 ]; // 1 is for term.
	if( buf )
	{
		if( mLine )
		{
			RqStrCopy(buf,mLine);
			delete[] mLine;
		}
		mSize = newSize;
		mLine = buf;
		return true;
	}
	return false;
}



