/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2007, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

 ////////                 
   //    //    ////////    This file is part of rqal, a.k.a. _Richard's Quality Library_ .
   //    //   //    //     Visit http://www.ranton.org for more information.
   //////     //    //     Email ranton@ranton.org for support or licensing.  
   //   //     //// //     
   //    //          //    Copyright (c) 2007 Richard Neal Anton
   //    //          //    All Rights Reserved.
  ///    //           //


#ifndef __RQ_LOG__
#define __RQ_LOG__

#ifndef __RQ_CONF__
#include "rq_conf.h"
#endif

#include <stdarg.h>
#ifdef __cplusplus
#include <iostream>
#include <sstream>
#endif

// These are the predefined logging levels,
// but the log mask is 32 bits so you have 25 bits of your own after the
// seven predefined values.
typedef int RqLogType;
const RqLogType RqLog_TRACE   = 1;
const RqLogType RqLog_DEBUG   = 2; // show debugging information
const RqLogType RqLog_INFO    = 4; // log informational stuff, very verbose most likely
const RqLogType RqLog_MEMORY  = 8; // show memory tracking information
const RqLogType RqLog_WARNING = 16; // log warnings
const RqLogType RqLog_ERROR	  = 32; // log errors and exception
const RqLogType RqLog_ASSERT  = 64; // failed assertions
const RqLogType RqLog_NONE    = 0; // log nothing
const RqLogType RqLog_ALL     = 0xFFFFFFFF; // log everything.


// these constants are to make it more convenient for you to assign the values for
// you own log message types.
const RqLogType RqLog_APP1    = 1<<7;
const RqLogType RqLog_APP2    = 1<<8;
const RqLogType RqLog_APP3    = 1<<9;
const RqLogType RqLog_APP4    = 1<<10;
const RqLogType RqLog_APP5    = 1<<11;
const RqLogType RqLog_APP6    = 1<<12;
const RqLogType RqLog_APP7    = 1<<13;
const RqLogType RqLog_APP8    = 1<<14;
const RqLogType RqLog_APP9    = 1<<15;
// If you need more than this you can figure it out yourself.

// INFO: Handling the log time flag is up to the callback, all of the builtin callbacks do so.

#ifdef __cplusplus
// set whether time value will be logged with each message; return is previous value; default is false.
bool RqSetLogTime(bool inLogTimeFlag);

// return value of log time flag.
bool RqGetLogTime(void);
#endif

// set callback for getting time value for log entries used if RqGetLogTime() == true
// default returns time(NULL)
typedef long (*RqTimeCallback)(void);
RqTimeCallback RqSetTimeCallback( RqTimeCallback inTimeCallback );

// Set log mask which determines which kinds of messages are actually output.
// returns prior value of mask.
RqLogType RqSetLogMask(RqLogType inMask);

// Function type for log output callback
typedef void (*RqLogCallback)(RqLogType kind, void *cbData, const char *fmt, va_list ap);

void *RqLogCallbackData(void);

// Set output callback for log messages
// changed to return old callback.
RqLogCallback RqSetLogCallback( RqLogCallback inCallback, void *cbData );

// -- Default time callback.
long RqDefaultTimeCallback(void);

// -- Default logging output callback.
void RqDefaultLogCallback(RqLogType kind, void *cbData, const char *fmt, va_list ap);

// -- Logging output callback for a stream
void RqStreamLogCallback(RqLogType kind, void *cbData, const char *fmt, va_list ap);

// -- Logging output for an fstream that flushes after each message
void RqFileLogCallback(RqLogType kind, void *cbData, const char *fmt, va_list ap);

// -- Logging output for capturing log, only meant for small amounts < 1k
void RqLogCaptureCallback(RqLogType kind, void *cbData, const char *fmt, va_list ap);
// get text captured by log while set to RqLogCaptureCallback
const char *RqLogCapture(void);
// reset log capture buffer.
void RqLogCaptureReset(void);

// output a message of a specified type
void RqLog(RqLogType kind, const char *fmt, ...);

// this is just a wrapper to call the proc
void RqLogV(RqLogType kind, const char *fmt, va_list ap);

// macro for safely outputting message.
#define _RqDoNotify(kind,fmt) { va_list ap; va_start( ap, fmt ); RqLogV(kind,fmt,ap); va_end(ap); }

// output a trace message with the file name and line number
#define RqTrace() RqLog(RqLog_TRACE,"file: %s line # %d",__FILE__,__LINE__)

void RqDebug(const char *fmt,...);
void RqInfo(const char *fmt, ...); 
void RqWarning(const char *fmt, ...); 
void RqLogError(const char *fmt, ...); 


// use this to temporarily suppress logging while executing a statement
// for those times when you just want it to shut the fuck up.

#ifdef __cplusplus
class RqLogInhibitor {
private:
	RqLogType mMask;
public:
	RqLogInhibitor(RqLogType inMask = RqLog_NONE)
		{
			mMask = RqSetLogMask(inMask);
		}
	~RqLogInhibitor()
		{
			RqSetLogMask(mMask);
		}
};

class RqLogStream {
public:
	RqLogStream( RqLogType inKind );
	~RqLogStream();

	RqLogStream& operator<< (const char *str );
	
	RqLogStream& operator<< (const std::string& sobj);

	RqLogStream& operator<< (bool val);
	RqLogStream& operator<< (char val);

	RqLogStream& operator<< (int val);
	RqLogStream& operator<< (unsigned val);
	RqLogStream& operator<< (long val);
	RqLogStream& operator<< (unsigned long val);
	RqLogStream& operator<< (double val);
	
	// could use long long / unsigned long long conversion
	// and a way to output hex conveniently.

private:
	enum {
		kConvertBufSize = 100
	};
	// this does the work.
	void Append(const char *str);
	// tries to grow, but deal with it if it doesn't.
	bool Grow(long minToGrow);

	// data members
	RqLogType mKind;
	char *mLine;
	int mPos, mSize;
};

inline RqLogStream rqlog( RqLogType inKind )
{
	return RqLogStream( inKind );
}

#define rqtrace rqlog(RqLog_TRACE)
#define rqdebug rqlog(RqLog_DEBUG)
#define rqinfo rqlog(RqLog_INFO)
#define rqmemory rqlog(RqLog_MEMORY)
#define rqwarning rqlog(RqLog_WARNING)
#define rqerror rqlog(RqLog_ERROR)

#define rqapp1 rqlog(RqLog_APP1)
#define rqapp2 rqlog(RqLog_APP2)
#define rqapp3 rqlog(RqLog_APP3)
#define rqapp4 rqlog(RqLog_APP4)
#define rqapp5 rqlog(RqLog_APP5)
#define rqapp6 rqlog(RqLog_APP6)
#define rqapp7 rqlog(RqLog_APP7)
#define rqapp8 rqlog(RqLog_APP8)
#define rqapp9 rqlog(RqLog_APP9)
#endif // #ifdef __cplusplus

#endif

