/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2007, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

 ////////                 
   //    //    ////////    This file is part of rqal, a.k.a. _Richard's Quality Library_ .
   //    //   //    //     Visit http://www.ranton.org for more information.
   //////     //    //     Email ranton@ranton.org for support or licensing.  
   //   //     //// //     
   //    //          //    Copyright (c) 2007 Richard Neal Anton
   //    //          //    All Rights Reserved.
  ///    //           //

#ifndef __RQ__
#include "rq.h"
#endif

// NOTE: rewrite the code in here so that allocated blocks
// have a header which points to their record.
// Actually, screw that, current way is safer but slower,
// if it's too slow then make it lookup via a hash.

// this makes it only include itself if rq_memory.h is included already,
// as in by the prefix file, and __BX_DEBUG__ is defined.
#ifdef RQ_MEM_DEBUG

#ifdef RQ_MACOS
// I'm not worried about the dependency on this since this
// code is only intended to run on our machines -RNA
#ifndef __THREADS__
#include <Threads.h>
#endif

// Don't want to get caught in our own macros
#undef NewPtr
#undef NewHandle
#undef DisposePtr
#undef DisposeHandle

#endif // RQ_MACOS

// Don't want to get caught in our own macros
#undef malloc
#undef free

#include <cstdio>
#include <exception>

// Metrowerks has cstdlib, but it doesn't have malloc/free
#if __MWERKS__
#include <stdlib.h>
#else
#include <cstdlib>
#endif

#include <stddef.h>

/* This is so we can turn on and off recording shit allocated by new
 * so that the stuff in this file can do its thing without being recorded */
static int sRecordNew = 1;

// true means we warn for bad memory accesses
static int sWarnAccess = 1;
struct RqMemAllocRecord
	{
	RqMemAllocRecord *next;	// next block in list
	RqMemAllocRecord *prev;	// previous block in list
	unsigned long serial;		// serial number for order checking
	void *ref;					// ptr/handle to block
	DebugBlockType type;		// type of block
	unsigned long size;			// size of the allocated block
	const char *file;			// file name of allocation
	unsigned long line;			// line number of allocation
	};

static const char kUnknownFileStr[] = "<unknown>";

static RqMemAllocRecord *gDebugRecList = NULL;
static unsigned long gDebugSerialNum = 0;
static RqMemAllocRecord *gDebugLastNew = NULL;
static RqUns8 *gDebugStackBase = NULL;

// this zaps memory that was just allocated.
static void ZapNew( void *ptr, long size )
{
	_RqMemSet16( ptr, 0xABCD, size>>1 );
}

// this zaps memory that was freed.
static void ZapFree( void *ptr, long size )
{
	_RqMemSet16( ptr, 0xCDEF, size>>1 );
}

RQ_MEM_EXTERN void RqMemReset( void )
	{
	RqMemAllocRecord *curRec,*nextRec;
	
	for ( curRec = gDebugRecList;curRec;curRec = nextRec )
		{
		nextRec = curRec->next;
		// DisposePtr( (Ptr)curRec );
		free((void *)curRec);
		}
	gDebugRecList = NULL;
	gDebugSerialNum = 0;
	}

void *operator new( size_t inSize )
	{
	if( sRecordNew )
		return RqMemAllocPointer( (unsigned long)inSize,NULL,0,kBlockNew );
	else
		return (void *)malloc( inSize );
	}
void *operator new[] (size_t inSize)
	{
	if( sRecordNew )
		return RqMemAllocPointer( (unsigned long)inSize,NULL,0,kBlockNew );
	else
		return (void *)malloc( inSize );
	}

void operator delete( void *inRef )
{
	if ( inRef )
	{
		if( sRecordNew )
			RqMemFreePointer( inRef, "", 0 );
		else
			free((void*)inRef);
	}
}
void operator delete[] ( void *inRef )
{
	if ( inRef )
	{
		if( sRecordNew )
			RqMemFreePointer( inRef, "", 0 );
		else
			free((void*)inRef);
	}
}

void _RqLabelNew(void *ref, char *inSrcFile,unsigned long inSrcLine)
{
	// store file/line information for a new operator call
	if( gDebugLastNew )
	{
		gDebugLastNew->file = inSrcFile;
		gDebugLastNew->line = inSrcLine;
		gDebugLastNew = NULL; // so we can't label it accidentally.
	}
	else
	{
		RqLog(RqLog_MEMORY,"Attempt to label unknown new block from %s:%lu",
			RqFile::GetFileNameFromFullPath(inSrcFile),inSrcLine);
	}
}

static RqMemAllocRecord* 
RqMemFindRef(void *ref)
{
	RqMemAllocRecord *curRec,*theRec = NULL;
	
	for ( curRec = gDebugRecList;curRec && !theRec;curRec = curRec->next )
		if ( curRec->ref == ref )
			theRec = curRec;

	return theRec;
}

// this only works with ptrs
static RqMemAllocRecord* 
RqMemFindContaining(void *ptr)
{
	RqMemAllocRecord* theRec = NULL;

	RqMemAllocRecord *curRec;
	
	for ( curRec = gDebugRecList; curRec; curRec = curRec->next )
	{
		RqUns8 *check;
#if defined(RQ_MACOS)
		if( curRec->type == kBlockHandle )
			check = (RqUns8 *)*(Handle)curRec->ref;
		else
#endif
			check = (RqUns8 *)curRec->ref;
		// has to be something other than a handle.
		//if( curRec->type != kBlockHandle && check <= ptr && ptr <= check + curRec->size )
		if( check <= ptr && ptr <= check + curRec->size )
		{
			theRec = curRec;
			break;
		}
	}
	return theRec;
}

// this function must be called at startup from main()
// so that values are properly initialized 
// for rq_memory
RQ_MEM_EXTERN void RqMemInit(void)
{
	int onstack;
	
	// If this is called from where it should be (inside main() or main entrypoint)
	// then this should be close to foolproof.

	//main objective is to store the stack pointer's initial value.
	gDebugStackBase = (RqUns8 *)&onstack;
}


RQ_MEM_EXTERN bool _RqMemOnStack(void *ptr,unsigned long inSize)
{
	// NOTE: we need to figure out how to do this
	// Probably just need to check if block is between stack pointer
	// and stack base pointer.
	RqUns8 here[1];
	RqUns8 *check = (RqUns8 *)ptr;

	// ATTN: This is absolute fucking voodoo and probably won't work.
	bool onStack = false;
	if( check >= here && check+inSize <= gDebugStackBase )
		onStack = true;
	return onStack;
}

// This checks that a block of memory specified by pointer/size is allocated, 
// it doesn't work for handles.
RQ_MEM_EXTERN int _RqCheckMemWork(void *ptr,unsigned long inSize)
{
	bool memoryGood = false;
	if( !inSize )
		return true; // they  aren't really accessing it.
	RqMemAllocRecord *theRec = RqMemFindContaining(ptr);
	if( theRec )
	{
		RqUns8 *check;
#if defined(RQ_MACOS)
		if( theRec->type == kBlockHandle )
			check = (RqUns8 *)*(Handle)theRec->ref;
		else
#endif
			check = (RqUns8 *)theRec->ref;
			
		if( ((RqUns8 *)ptr) + inSize <= check + theRec->size )
			memoryGood = true;
	}
	if( !memoryGood )
		memoryGood =_RqMemOnStack(ptr,inSize);
		
	return memoryGood;
}

// This uses _RqCheckMemWork to verify ptr/inSize then handles failure if needed.
RQ_MEM_EXTERN void _RqCheckMem(void *ptr,unsigned long inSize,const char *inSrcFile,unsigned long inSrcLine)
{
	// find containing block and check if it is large enough.
	bool memoryGood = _RqCheckMemWork(ptr,inSize) != 0;

	// Log it
	// we were asserting, but there are too many cases where things were
	// allocated by some outside library or the OS and there is no real
	// way to have marked them.
	if( !memoryGood && sWarnAccess)
		RqLog(RqLog_MEMORY,"Bad memory access attempted.");
}

static RqMemAllocRecord *
RqNewMemRecord(void *ref, unsigned long inSize, const char *inSrcFile, unsigned long inSrcLine, DebugBlockType inType)
{
	RqMemAllocRecord* newRec = (RqMemAllocRecord *)malloc( sizeof(RqMemAllocRecord) );
	if( newRec)
	{
		newRec->serial = gDebugSerialNum++;
		newRec->type = inType;
		if ( inSrcFile )
			{
			newRec->file = inSrcFile;
			newRec->line = inSrcLine;
			}
		else
			{
				if(inType == kBlockNew)
					gDebugLastNew = newRec;
			newRec->file = NULL;
			newRec->line = 0;
			}
		newRec->size = inSize;
		newRec->ref = ref;
		newRec->prev = NULL;
		newRec->next = NULL;
	}
	return newRec;
}

// remove and free a record.
static void
RqMemRemoveRecord(RqMemAllocRecord *theRec)
{
	if ( theRec->prev )
		theRec->prev->next = theRec->next;
	else
		gDebugRecList = theRec->next;
	if ( theRec->next )
		theRec->next->prev = theRec->prev;
	free( (void*)theRec );
}


// this declares staticly defined objects, or other things such as GlobalAlloc, locked handles, etc.
RQ_MEM_EXTERN void _RqDeclareMem(void *ptr,unsigned long inSize,const char *inSrcFile,unsigned long inSrcLine)
{
	RqMemAllocRecord *newRec = RqNewMemRecord(ptr,inSize,inSrcFile,inSrcLine,kBlockDeclared);

	if( !newRec )
		RqLog(RqLog_MEMORY,"_RqDeclareMem: Allocation failure");
	else
	{
		// Add it to the list.
		newRec->next = gDebugRecList;
		if ( gDebugRecList )
			gDebugRecList->prev = newRec;
		gDebugRecList = newRec;
	}
}

// remove record for a block that is being freed or whatever.
RQ_MEM_EXTERN void _RqUndeclareMem( void *ptr, char *inSrcFile,unsigned long inSrcLine )
{
	RqMemAllocRecord *theRec = RqMemFindRef(ptr);
	if ( theRec )
		RqMemRemoveRecord(theRec);
	else
		RqLog(RqLog_DEBUG,"Undeclaring unknown ptr from %s:%lu",inSrcFile,inSrcLine);
}

RQ_MEM_EXTERN void* 
RqMemAllocPointer( unsigned long inSize,const char *inSrcFile,unsigned long inSrcLine,DebugBlockType inType )
	{
	RqMemAllocRecord *newRec = NULL;
	void* result;

	result = malloc( inSize );
	
	if( result )
	{
		// fuck it up
		ZapNew( (void *)result, inSize );
		// create the record for  it
		newRec = RqNewMemRecord(result,inSize,inSrcFile,inSrcLine,inType);
	}
	if( !newRec )
	{
		RqLog(RqLog_MEMORY,"RqMemAllocPointer: Allocation failed");
		free(result);
		return NULL;
	}

	// store the record in the list
	newRec->next = gDebugRecList;
	if ( gDebugRecList )
		gDebugRecList->prev = newRec;
	gDebugRecList = newRec;
	return result;
	}


RQ_MEM_EXTERN void RqMemFreePointer( void* ref, char *inSrcFile,unsigned long inSrcLine )
	{
	RqMemAllocRecord *theRec = RqMemFindRef(ref);
	unsigned long sz = 0;
	if ( theRec )
		{
		sz = theRec->size;
		RqMemRemoveRecord(theRec);
		}
	else
	{
		// log that an unnaccounted for block is being freed.
		RqLog(RqLog_DEBUG,"Freeing unknown ptr from %s:%lu",inSrcFile,inSrcLine);
	}

	RqAssert( ref );
	ZapFree( ref, sz );
	free( (void*)ref );
	}

#ifdef RQ_MACOS
RQ_MEM_EXTERN Handle RqMemAllocHandle( unsigned long inSize,const char *inSrcFile,unsigned long inSrcLine )
	{
	RqMemAllocRecord *newRec = NULL;
	Handle result;
	
	result = NewHandle( inSize );
	if ( result != NULL && MemError() == noErr )		
		newRec = RqNewMemRecord(result,inSize,inSrcFile,inSrcLine,kBlockHandle);
	if(!newRec)
	{
		RqLog(RqLog_MEMORY,"RqMemAllocHandle: Allocation failed");
		DisposeHandle(result);
		return NULL;
	}

	newRec->next = gDebugRecList;
	if ( gDebugRecList )
		gDebugRecList->prev = newRec;
	gDebugRecList = newRec;
	return result;
	}

RQ_MEM_EXTERN void RqMemFreeHandle( Handle ref, const char *inSrcFile,unsigned long inSrcLine )
	{
	RqMemAllocRecord* theRec = RqMemFindRef((void *)ref);

	if ( theRec )
		{
		if ( theRec->prev )
			theRec->prev->next = theRec->next;
		  else
			gDebugRecList = theRec->next;
		if ( theRec->next )
			theRec->next->prev = theRec->prev;
		free( (void *)theRec );
		}
	else
	{
		// log that an unnaccounted for block is being freed.
		RqLog(RqLog_DEBUG,"Freeing unknown handle from %s:%lu",inSrcFile,inSrcLine);
	}
	DisposeHandle( ref );
	}

RQ_MEM_EXTERN Handle RqMemAllocHandleClear( unsigned long inSize,const char *inSrcFile,unsigned long inSrcLine )
{
	Handle temp = RqMemAllocHandle( inSize,inSrcFile, inSrcLine );
	if( temp )
	{
		HLock( temp );
		_RqMemSet8( (void *)*temp, 0, inSize );
		HUnlock( temp );
	}
	return temp;
}
#endif



RQ_MEM_EXTERN void *RqMemAllocPointerClear( unsigned long inSize,const char *inSrcFile,unsigned long inSrcLine,DebugBlockType inType )
{
	void* temp = RqMemAllocPointer( inSize,inSrcFile, inSrcLine, inType );
	if( temp )
		_RqMemSet8( temp, 0, inSize );
	return temp;
}



// this does the work for RqMemSummary and RqMemDumpBlocks
static void RqMemDumpWork( const char *inComment, bool inSummarize )
	{
	// Very first thing, turn off the new/delete shit
	// so that we don't cause bogus leak information while we are using BxString
	
	// This may be important in threaded stuff.
	
#ifdef RQ_MACOS
	ThreadBeginCritical();
#endif

	sRecordNew = false;
	try
	{
		/* screw this comment shit, I'm not even sure what the hell it's point was */
		if ( inComment )
			RqLog(RqLog_MEMORY,"(%s)",inComment);
		
		// Traverse the list to the end, accumulating statistics on the way
		unsigned long recCount=0, recSize = 0;

		// don't count statically allocated blocks in any of the statistics.
		RqMemAllocRecord* curRec = gDebugRecList;
		while ( curRec && curRec->next )
		{
			if( curRec->type != kBlockDeclared )
			{
				recCount++;
				recSize += curRec->size;
			}
			curRec = curRec->next;
		}
		// Get that one last record
		if ( curRec && curRec->type != kBlockDeclared)
		{
			recCount++;
			recSize += curRec->size;
		}
		
		if ( recCount > 0 )
		{
			// Output our header
			if( !inSummarize )
			{
				RqLog(RqLog_MEMORY,"%8s%30s%6s%10s%8s","Serial","Filename","Line","Type","Size" );
				RqLog(RqLog_MEMORY,"--------------------------------------------------------------" );
			}
			//fprintf( f,"%8s%30s%6s%10s%8s\n","Serial","Filename","Line","Type","Size" );
			//fprintf( f,"--------------------------------------------------------------\n" );
			//str = 62 spaces + new line

			// Traverse the list in the reverse order, outputting the records
			while ( curRec )
				{
				// don't print static blocks.
				if( curRec->type == kBlockDeclared )
				{
					curRec = curRec->prev;
					continue;
				}

				const char *typeStr = "";
				switch ( curRec->type )
					{
					case kBlockPointer:
						typeStr = "Pointer";
						break;
					case kBlockHandle:
						typeStr = "Handle";
						break;
					case kBlockNew:
						typeStr = "New";
						break;
					case kBlockDeclared:
						typeStr = "Declared";
						break;
					}
				const char *fileStr = kUnknownFileStr;
				// don't log the entire pathname or it will get out of hand.
				if( curRec->file )
					fileStr = RqFile::GetFileNameFromFullPath( curRec->file );

				if( !inSummarize )
				{
					RqLog(RqLog_MEMORY,"%8lu%30s%6lu%10s%8lu",
						curRec->serial,fileStr,curRec->line,typeStr,curRec->size);
				}

				curRec = curRec->prev;
				}
			
			if( !inSummarize )
			{
				RqLog(RqLog_MEMORY,"--------------------------------------------------------------");
			}

			if ( recCount == 1 )
				RqLog(RqLog_MEMORY,"1 block detected, totalling %lu bytes.",recSize );
			else
				RqLog(RqLog_MEMORY,"%lu blocks detected, totalling %lu bytes.",recCount,recSize );
		
		}
		else
		{
			RqLog(RqLog_MEMORY,"No blocks detected");
			//fprintf( f,"No blocks detected\n" );
		}
	} 
	catch(std::exception& e)
	{
		RqLog(RqLog_MEMORY,"%s",e.what());
	}
	catch(RqError& rqe)
	{
		RqLog(RqLog_MEMORY,"%s (%d) from %s:%d",rqe.Message(),rqe.Value(),rqe.File(),rqe.Line());
	}
	catch(... ) 
	{
		// we had an error.
		// just fall through to cleanup.
	}
	
	// Turn this back on.
	sRecordNew = true;

#ifdef RQ_MACOS
	ThreadEndCritical();
#endif

	}	

RQ_MEM_EXTERN void RqMemDisableAccessWarnings( void )
{
	sWarnAccess = false;
}

// Log a summary of how much memory is outstanding.
RQ_MEM_EXTERN void RqMemSummary( const char *inComment )
{
	RqMemDumpWork( inComment, true );
}

// This logs information about all of the outstanding blocks.
RQ_MEM_EXTERN void RqMemDumpBlocks( const char *inComment )
{
	RqMemDumpWork( inComment, false );
}


#endif // RQ_MEM_DEBUG

// These are memory and string utility routines that are
// defined either to standard ones or to debugging versions
RQ_MEM_EXTERN void _RqMemCopy(void *dst,const void *src,size_t size)
{
	// check that dst[size] and src[size] are in an allocated block.
	RqCheckMem(dst,size);
	RqCheckMem(src,size);
	RqUns8 *in,*out;
	
	in = (RqUns8 *)src;
	out = (RqUns8 *)dst;
	while(size--)
		*out++ = *in++;
}
RQ_MEM_EXTERN void _RqMemSet8(void *dst,int val,size_t size)
{
	RqCheckMem(dst,size);

	RqUns8* out, byte = (RqUns8)val;

	out = (RqUns8 *)dst;
	while(size--)
		*out++ = byte;
}
RQ_MEM_EXTERN int _RqMemCompare(const void *dst,const void *src,size_t size)
{
	// check that dst, and src are both inside allocated memory block.
	RqCheckMem(dst,size);
	RqCheckMem(src,size);

	register int diff;
	const RqUns8 *p1 = (RqUns8 *)dst, *p2 = (RqUns8 *)src;
	
	while(size--)
	{
		diff = *p1++;
		diff -= (int)*p2++;
		if(diff)
			break;
	}
	return diff;
}
// This one has no corresponding equivalent in stdlib so it's not a macro
RQ_MEM_EXTERN void _RqMemSet16(void *dst,RqUns16 val,size_t inCount)
{
	RqAssert(dst);
	RqUns16 *out;

	out = (RqUns16 *)dst;
	while(inCount--)
		*out++ = val;
}
// same with this one
RQ_MEM_EXTERN void _RqMemSet32(void *dst,RqUns32 val,size_t inCount)
{
	RqAssert(dst);
	RqUns32 *out;

	out = (RqUns32 *)dst;
	while(inCount--)
		*out++ = val;
}

RQ_MEM_EXTERN void RqMemSet16(void *dst,RqUns16 val,size_t inCount)
{
	RqCheckMem(dst,inCount*2);
	_RqMemSet16(dst,val,inCount);
}
RQ_MEM_EXTERN void RqMemSet32(void *dst,RqUns32 val,size_t inCount)
{
	RqCheckMem(dst,inCount*4);
	_RqMemSet32(dst,val,inCount);
}

RQ_MEM_EXTERN long _RqStrLen(const char *str)
{
	// NOTE: check that str falls somewhere in allocated memory,
	// and how much is allocated.
	RqAssert(str);

	long len = 0;
	while( *str++ )
		len++;
	return len;
}

// compare two strings, in style of strcmp(),
// return < 0 means inA < inB, etc...
RQ_MEM_EXTERN int _RqStrCmp(const char *inA, const char *inB)
{
	RqAssert( inA && inB );

	int diff = 0; // this was missing this, which is a bad bug.
	
	// if both empty this will explode.
	if( !*inA && !*inB )
		return 0;

	// loop until different or both end(if one ends they'll be different)
	while( (diff=*inA++ - *inB++) == 0 && *inA )
		;
	return diff;
}

// copy one string to another, no bounds checking is done
RQ_MEM_EXTERN void _RqStrCopy(char *outStr, const char *inStr)
{
	RqAssert( outStr && inStr );
	while(*inStr)
		*outStr++ = *inStr++;
	*outStr++ = 0; // and terminator
}


// cat two strings together
RQ_MEM_EXTERN void _RqStrCat( char *dst, const char *src )
{
	RqUns32 clen = RqStrLen(dst);
	RqUns32 alen = RqStrLen(src);
	RqMemCopy(dst+clen,src,alen);
	dst[clen+alen] = 0;
}

// compare 2 strings, ignoring case
RQ_MEM_EXTERN RqInt32 _RqStrCaseCmp( const char *a, const char *b )
{
	RqInt32 diff = 0;
	char ac, bc;

	RqAssert( a && b );

	for(;;)
	{
		ac = *a;
		if( ac >= 'a' && ac <= 'z' )
			ac += 'A' - 'a';

		bc = *b;
		if( bc >= 'a' && bc <= 'z' )
			bc += 'A' - 'a';

		diff = ac - bc;
		if( diff || !ac || !bc ) break;
		a++;
		b++;
	}

	return diff;
}

// compare 2 strings, ignoring case, up to a maximum length
RQ_MEM_EXTERN RqInt32 _RqStrCaseNCmp( const char *a, const char *b, RqUns32 n )
{
	RqInt32 diff = 0;
	char ac, bc;
	RqUns32 index = 0;

	RqAssert( a && b );

	for(index=0;index<n;index++)
	{
		ac = *a;
		if( ac >= 'a' && ac <= 'z' )
			ac += 'A' - 'a';

		bc = *b;
		if( bc >= 'a' && bc <= 'z' )
			bc += 'A' - 'a';

		diff = ac - bc;
		if( diff || !ac || !bc ) break;
		a++;
		b++;
	}

	return diff;
}
